function [ x, out, opts ] = tfocs_N83( smoothF, affineF, projectorF, x0, opts )
% TFOCS_N83 Nesterov's 1983 accelerated method; also by Beck and Teboulle 2005 (FISTA).
% [ x, out, opts ] = tfocs_N83( smoothF, affineF, nonsmoothF, x0, opts )
%   Implements Nesterov's 1983 method.
%   A variety of calling sequences are supported; type 
%      help tfocs_help
%   for a full explanation.

% Initialization
alg = 'N83';
algorithm = 'Nesterov''s 1983 single-projection method';
alpha = 0; beta = 0; mu = 0; L = 0; % Do not remove: necessary because of a MATLAB quirk
if(~isfield(opts, 'max_time'))
    opts.max_time = inf;
end
tbound = opts.max_time;
tfocs_initialize
if nargin == 0, return; end
tic
outtimes = 0;
[n, k] = size(x);
temp = z;
clear('z');
z.U = temp;
z.D = eye(size(z.U, 2));
nf = 0;
ng = 0;
while true,
      x_old =   x;
      z_old =   z;
    
    % The backtracking loop
    L_old      = L;
   	L          = L * alpha;
    theta_old  = theta;
    times = 0;
    while true,
        times = times + 1;
        % Acceleration
        theta = advance_theta( theta_old, L, L_old );
        % Next iterate
        if theta < 1,
            y = low_rank_addition(( 1 - theta ), x_old , theta, z_old, k);
            f_y = Inf;
            g_y = [];
        end
        % Compute function values
        if isempty( g_y ),
            [ f_y, g_y ] = apply_smooth( y ); 
            nf = nf + 1;
            ng = ng + 1;
        end

        % Standard gradient
        x = low_rank_add_gfv(1, y, - (1/L), g_y, k);
        f_x = Inf;
        g_x = [];
        % TFOCS_BACKTRACK
        % Quick exit for no backtracking
        if beta >= 1, break; end % SRB changing == to >=
        % Quick exit if no progress made
        xy = low_rank_substraction(1, x, -1, y);
        temp = xy.U' * xy.U;
        xy_sq = real(trace(xy.D * temp * xy.D' * temp));
        if xy_sq == 0, localL = Inf; 
            break; 
        end

        % Compute Lipschitz estimate
        if backtrack_simple,
            if isinf( f_x ),
                f_x = apply_smooth( x );
                nf = nf + 1;
            end
            first_order = inpro_gf_v(g_y, xy);
            q_x = f_y + first_order + 0.5 * L * xy_sq;

            localL = L + 2 * max( f_x - q_x, 0 ) / xy_sq;
            backtrack_simple = abs( f_y - f_x ) >= backtrack_tol * max( abs( f_x ), abs( f_y ) );
%             fprintf('\t f_x - f_y:%e, inpro:%e, second_order:%e\n', f_x - f_y, first_order, 0.5 * L * xy_sq);%%---
%             fprintf('\t L:%e, localL:%e, f_x:%e, q_x:%e, f_y:%e\n', L, localL, f_x, q_x, f_y);%%---
%             fprintf('\t diff_f:%e, fx:%e, fy:%e, backtrack_simple:%d, backtrack_tol:%e\n', abs( f_y - f_x ), f_x, f_y, backtrack_simple, backtrack_tol);%%---
        else
            if isempty( g_x ),
                [ f_x, g_x ] = apply_smooth( x );
                nf = nf + 1;
                ng = ng + 1;
            end
            localL = 2 * inpro_gf1addgf2_v(1, g_x, -1, g_y, xy) / xy_sq;
        end

        % Exit if Lipschitz criterion satisfied, or if we hit Lexact
        backtrack_steps = backtrack_steps + 1;
        if localL <= L || L >= Lexact,
            break; 
        end
        
        if ~isinf( localL ), 
            L = min( Lexact, localL );
        elseif isinf( localL ), localL = L; end
        L = min( Lexact, max( localL, L / beta ) );
        % TFOCS v1.3 by Stephen Becker, Emmanuel Candes, and Michael Grant.
        % Copyright 2013 California Institute of Technology and CVX Research.
        % See the file LICENSE for full license information.

    end
%     fprintf('out-iter:%d, in-iter:%d, f:%e \n', outtimes, times, f_x);
    outtimes = outtimes + 1;

    % Collect data, evaluate stopping criteria, and print status
    tfocs_iterate_PR

    % Reversed z update. This step is skipped if restart occurs
    if theta == 1,
        z.U   =   x;
        z.D = eye(size(z.U, 2));
    else
        z = low_rank_substraction((1/theta), x , ( 1 - 1/theta ), x_old);
    end    
end
t = toc;
fprintf('timecost:%f \n', t);
opts.timecost = t;
opts.iter = outtimes;
opts.nf = nf;
opts.ng = ng;
opts.f_final = f_x;

% Final processing
tfocs_cleanup

% TFOCS v1.3 by Stephen Becker, Emmanuel Candes, and Michael Grant.
% Copyright 2013 California Institute of Technology and CVX Research.
% See the file LICENSE for full license information.

function output = low_rank_addition(a, x, b, y, k)
    if(a < 0 || b < 0)
        error('a and b must be nonnegative!');
    end
    [n, k1] = size(x);
    k2 = size(y, 2);
    Z = [x, y.U];
    [Q, R] = qr(Z, 0);
    M = R * diag([ones(k1, 1) * a; diag(y.D) * b]) * R';
    [U, D] = eig(M);
    D = real(D);
    [d, idx] = sort(diag(D), 'descend');
    output = Q * U(:, idx(1:k)) * diag(sqrt(max(d(1:k), 0)));
return


function output = low_rank_substraction(a, x, b, y)
    if(a < 0 || b > 0)
        error('a must be nonnegative and b must be nonpositive');
    end
    [n, k1] = size(x);
    k2 = size(y, 2);
    Z = [x, y];
    [Q, R] = qr(Z, 0);
    M = R * diag([ones(k1, 1) * a; ones(k2, 1) * b]) * R';
    [U, D] = eig(M);
    D = real(D);
    output.U = Q * U;
    output.D = D;
return

function output = gf2norm(gf, n, k)
    func = @(v)gf(v);
    opts.isreal = false;
    opts.issym = 1;
    [U, D] = eigs(func, n, 1, 'lm', opts);
    output = abs(D);
return


function output = low_rank_add_gfv(a, x, b, gf, k)
    n = size(x, 1);
    func = @(v) a * x * (x' * v) + b * gf(v);
    opts.isreal = false;
    opts.issym = 1;
    [U, D] = eigs(func, n, k, 'lr', opts);
    D = real(D);
    D = max(D, 0);
    output = U * sqrt(D);
return

function output = inpro_gf_v(gf, eta)
    temp = gf(eta.U) * eta.D';
    output = real(eta.U(:)' * temp(:));
return

function output = inpro_gf1addgf2_v(a, gf1, b, gf2, v)
    temp = (a * gf1(v.U) + b * gf2(v.U)) * v.D;
    output = real(v.U(:)' * temp(:));
return
